/*
 ATtiny25 Tiny Temperature Logger (A25TTL) - works with ATtiny13 too
 Version: 2.0
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 12/11/2012
 Last Modified: 21/09/2014

 */

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#ifndef boolean
	typedef uint8_t boolean;
#endif
#ifndef bool
	typedef uint8_t bool;
#endif 
#ifndef byte
	typedef uint8_t byte;
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define LOW 0
#define HIGH 1
#define false 0
#define true 1

// Watchdog timeouts to values
#define T16MS 0
#define T32MS 1
#define T64MS 2
#define T128MS 3
#define T250MS 4
#define T500MS 5
#define T1S 6
#define T2S 7
#define T4S 8
#define T8S 9

#define blankEeprom2byte 65535 // 255 high and 255 low byte means the 2 bytes are unprogrammed

#define NOREPEATSTART 0
#define REPEATSTART 1

#define I2C_DELAY_USEC 5
#define I2C_READ 1
#define I2C_WRITE 0

// Soft I2c Library used from http://forums.adafruit.com/viewtopic.php?f=25&t=13722
// Initialize SCL/SDA pin and set the bus high
void SoftI2cMasterInit(void) {
	DDRB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SDA_PIN);
	DDRB |= (1<<TWI_SCL_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
}

// De-initialize SCL/SDA pins and set the bus low
void SoftI2cMasterDeInit(void) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	DDRB &= ~(1<<TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	// Don't re-set the clock to an input
}

// Read a byte from I2C and send Ack if more reads follow else Nak to terminate read
uint8_t SoftI2cMasterRead(uint8_t last) {
	uint8_t b = 0;
	// Make sure pull-up enabled
	PORTB |= (1<<TWI_SDA_PIN);
	DDRB &= ~(1<<TWI_SDA_PIN);
	// Read byte
	for (uint8_t i = 0; i < 8; i++) {
		// Don't change this loop unless you verify the change with a scope
		b <<= 1;
		_delay_us(I2C_DELAY_USEC);
		PORTB |= (1<<TWI_SCL_PIN);
		if (bit_is_set(PINB, TWI_SDA_PIN)) b |= 1;
		PORTB &= ~(1<<TWI_SCL_PIN);
	}
	// Send Ack or Nak
	DDRB |= (1<<TWI_SDA_PIN);
	if (last) { 
		PORTB |= (1<<TWI_SDA_PIN); 
	}
	else { 
		PORTB &= ~(1<<TWI_SDA_PIN);
	}
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB &= ~(1<<TWI_SCL_PIN);
	PORTB &= ~(1<<TWI_SDA_PIN);
	return b;
}

// Write a byte to I2C
bool SoftI2cMasterWrite(uint8_t data) {
	// Write byte
	for (uint8_t m = 0x80; m != 0; m >>= 1) {
		// Don't change this loop unless you verify the change with a scope
		if (m & data) { 
			PORTB |= (1<<TWI_SDA_PIN); 
		}
		else { 
			PORTB &= ~(1<<TWI_SDA_PIN); 
		}
		PORTB |= (1<<TWI_SCL_PIN);
		_delay_us(I2C_DELAY_USEC);
		PORTB &= ~(1<<TWI_SCL_PIN);
	}
	// get Ack or Nak
	DDRB &= ~(1<<TWI_SDA_PIN);
	// Enable pullup
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	uint8_t rtn = bit_is_set(PINB, TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	PORTB &= ~(1<<TWI_SDA_PIN);
	DDRB |= (1<<TWI_SDA_PIN);
	return rtn == 0;
}

// Issue a start condition
bool SoftI2cMasterStart(uint8_t addressRW) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB &= ~(1<<TWI_SCL_PIN);
	return SoftI2cMasterWrite(addressRW);
}

// Issue a restart condition
bool SoftI2cMasterRestart(uint8_t addressRW) {
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	return SoftI2cMasterStart(addressRW);
}

// Issue a stop condition
void SoftI2cMasterStop(void) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB |= (1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
}

// Issue the device start and the address to write
bool SoftI2cMasterAddress(uint8_t deviceAddr, uint16_t addressRW, bool repeatedStart) {
	if (!SoftI2cMasterStart((deviceAddr<<1) | I2C_WRITE)) return false;
	if (!SoftI2cMasterWrite((addressRW >> 8))) return false; // MSB
	if (!SoftI2cMasterWrite((addressRW & 0xFF))) return false; // LSB
	if (repeatedStart) { if (!SoftI2cMasterRestart((deviceAddr<<1) | I2C_READ)) return false; }
	
	return true;
}

// Read 1 byte from the EEPROM device and return it
uint8_t soft_i2c_eeprom_read_byte(uint8_t deviceAddr, uint16_t readAddress) {
	uint8_t byteRead = 0;
	
	// Issue start condition, send device address, issue repeated start
	SoftI2cMasterAddress(deviceAddr, readAddress, REPEATSTART);
	
	// Read the byte
	byteRead = SoftI2cMasterRead(1);
	
	// Issue a stop condition
	SoftI2cMasterStop();
	
	return byteRead;
}

// Write 1 byte to the EEPROM
bool soft_i2c_eeprom_write_byte(uint8_t deviceAddr, uint16_t writeAddress, byte writeByte) {
	
	// Issue start condition and send device address
	SoftI2cMasterAddress(deviceAddr, writeAddress, NOREPEATSTART);
	
	// Write the byte
	if (!SoftI2cMasterWrite(writeByte)) return false;
	
	// Issue a stop condition
	SoftI2cMasterStop();
	
	// Delay 6ms for the write to complete. This value depends on the EEPROM you use so check your EEPROM's datasheet.
	_delay_ms(6);
	
	return true;
}

// Put the ATtiny in power down sleep mode and switch off the ADC too
void system_sleep(void) {
	cbi(ADCSRA,ADEN); // Switch Analog to Digital converter Off
	set_sleep_mode(SLEEP_MODE_PWR_DOWN); // Set sleep mode
	sleep_mode(); // System sleeps here
	sbi(ADCSRA,ADEN);  // Switch Analog to Digital converter On
}

// Watchdog timeout values
// 0=16ms, 1=32ms, 2=64ms, 3=128ms, 4=250ms, 5=500ms, 6=1sec, 7=2sec, 8=4sec, 9=8sec
// Used from http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
void setup_watchdog(byte ii) {
	byte bb;
	if (ii > 9 ) ii=9;
	bb=ii & 7;
	if (ii > 7) bb|= (1<<5);
	bb|= (1<<WDCE);
	
	MCUSR &= ~(1<<WDRF);
	// Start timed sequence
	WDTCR |= (1<<WDCE) | (1<<WDE);
	// Set new watchdog timeout value
	WDTCR = bb;
	WDTCR |= _BV(WDTIE);
}

// Turn off the Watchdog
void turnOffwatchdog(void) {
	cli(); // Turn interrupts off
	MCUSR &= ~(1<<WDRF);
	WDTCR |= (1<<WDCE) | (1<<WDE);
	WDTCR = 0;
	sei(); // Turn interrupts back on
}

// Watchdog sleep function - combines all the above functions into one
void watchdogSleep(byte timeout) { 
	setup_watchdog(timeout);
	system_sleep();
	turnOffwatchdog();
}

// Read the ADC
int analogRead(uint8_t pin) {
	uint8_t low, high;
	
	// Set the analog reference
	ADMUX = pin & 0x3f;
	
	// Set ADC sleep mode and wake up once the ADC has finished
	set_sleep_mode(SLEEP_MODE_ADC);
	sleep_mode();
	
	// Read low byte first otherwise reading the high byte first would discard the low byte
	low = ADCL;
	high = ADCH;
	
	// Combine the two bytes
	return (high << 8) | low;
}

// Setup the ATtiny
void setup(void) {
	// Set Outputs
	DDRB |= (1<<tempPin);
	
	// Turn on pull-up resistors on other ports to save power
	PORTB |= (1<<PB0);
	
	// Turn on interrupts
	sei();
	
	// Set pull-up resistors I2C SCL and SDA pins (some EEPROMs draw 100-300uA if SCL/SDA lines aren't set up)
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	
	// Wait a little bit for things to settle
	watchdogSleep(T1S);
	
	// Check if we powered up from the USB port as the SCL would now be low due to the AT25TTL Reader pulling it to ground
	// If it's low, sleep forver as the reader will be accessing the EEPROM now
	if (!(PINB & (1<<TWI_SCL_PIN))) {
		system_sleep(); // Sleep forever
	}
	
	// Turn off pull up resistors as we are the I2C master now and will be driving them directly
	PORTB &= ~(1<<TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	
	// Initialse I2C
	SoftI2cMasterInit();
	
	// Wait 8 seconds before logging to make sure the battery wasn't disconnected and re-connected when being removed or installed
	watchdogSleep(T8S);
	
	// Set ADC prescale factor to 8, so we have 1 MHz clock / 8 = 125 KHz, inside the desired 50-200 KHz range.
	sbi(ADCSRA, ADPS1);
	sbi(ADCSRA, ADPS0);
	
	// Enable ADC conversions
	sbi(ADCSRA, ADEN);
	
	// Turn on ADC interrupt
	sbi(ADCSRA, ADIE);
}
